from yta_general_utils.downloader.google_drive import GoogleDriveResource
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.file.checker import file_exists
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.path import create_file_abspath


def get_resource(google_drive_url: str, output_filename: str = None):
    """
    This method has been built to be used internally to obtain
    the resources this software needs. This method will look
    for the 'output_filename' if provided, and if it exist
    locally stored, it will return that filename. If it doesn't
    exist or the 'output_filename' is not provided, it will 
    download the resource from the given 'google_drive_url',
    store it locally and return it.

    The first time a resource is downloaded from the given
    'google_drive_url', if 'output_filename' provided, will
    be stored locally with that filename. Next time, as it 
    will be found locally, it will be returned and we will
    have not to wait for the download.

    This method has been created to optimize the resource
    getting process.

    @param
        **google_drive_url**
        The Google Drive url in which we can found (and
        download) the resource.

    @param
        **output_filename**
        The filename of the resource we want to get, that
        must be a relative path from the current project
        abspath.
    """
    if not google_drive_url:
        # TODO: Maybe raise exception (?)
        return None
    
    output_abspath = create_temp_filename('tmp_resource.mp3')
    
    if output_filename:
        # We try to find it and return or we create the folder if doesn't exist
        output_abspath = get_project_abspath() + output_filename

        if file_exists(output_abspath):
            return output_abspath
        
        # We force to create all folders
        create_file_abspath(output_abspath)
    
    # We download the file and store it locally
    return GoogleDriveResource(google_drive_url).download(output_abspath)


from yta_multimedia.greenscreen.utils import get_greenscreen_areas_details
from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from yta_multimedia.greenscreen.classes.greenscreen_area_details import GreenscreenAreaDetails
from yta_multimedia.video.frames import VideoFrameExtractor
from yta_multimedia.greenscreen.enums import GreenscreenType
from yta_multimedia.greenscreen.utils import GREENSCREENS_FOLDER
from yta_multimedia.resources import get_resource
from yta_general_utils.downloader.google_drive import GoogleDriveResource
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.checker.url import is_google_drive_url


def get_greenscreen_details(greenscreen_filename_or_google_drive_url: str, type: GreenscreenType):
    """
    Method to obtain greenscreen area and details that must be
    used by ImageGreenscreen and VideoGreenscreen to automatically
    detect greenscreens in their resources.
    """
    if not greenscreen_filename_or_google_drive_url:
        return None
    
    # TODO: Check that 'type' is GreenscreenType type

    # We will need the resource filename or google drive url and
    # the image to extract the data from
    RESOURCE_FILENAME = greenscreen_filename_or_google_drive_url
    TMP_FILENAME = greenscreen_filename_or_google_drive_url

    if type == GreenscreenType.IMAGE:
        # We have the final resource and the image to extract data
        if is_google_drive_url(greenscreen_filename_or_google_drive_url):
            google_drive_id = GoogleDriveResource(greenscreen_filename_or_google_drive_url).id
            filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.png'
            TMP_FILENAME = get_resource(greenscreen_filename_or_google_drive_url, filename)
    elif type == GreenscreenType.VIDEO:
        if is_google_drive_url(greenscreen_filename_or_google_drive_url):
            google_drive_id = GoogleDriveResource(greenscreen_filename_or_google_drive_url).id
            filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4'
            RESOURCE_FILENAME = get_resource(greenscreen_filename_or_google_drive_url, filename)
        TMP_FILENAME = create_temp_filename('tmp_gs_autodetect.png')
        VideoFrameExtractor.get_frame_by_frame_number(RESOURCE_FILENAME, 0, TMP_FILENAME)
    else:
        raise Exception(f'The greenscreen type "{type}" is not a valid image or video file.')

    green_rgb_color, similar_greens, regions = get_greenscreen_areas_details(TMP_FILENAME)

    if not green_rgb_color:
        raise Exception('No automatic greenscreen detected in "' + greenscreen_filename_or_google_drive_url + '". Aborting.')

    greenscreen_areas = []
    for region in regions:
        greenscreen_areas.append(GreenscreenAreaDetails(
            rgb_color = green_rgb_color,
            similar_greens = similar_greens,
            # TODO: Rename this (?)
            upper_left_pixel = region.top_left,
            lower_right_pixel = region.bottom_right,
            frames = None # TODO: Implement a way of handling frames
        ))

    greenscreen_filename_or_google_drive_url = GreenscreenDetails(
        greenscreen_areas = greenscreen_areas,
        filename_or_google_drive_url = RESOURCE_FILENAME,
        type = type
    )

    return greenscreen_filename_or_google_drive_url


from yta_multimedia.resources import get_resource
from yta_multimedia.video.frames import VideoFrameExtractor
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.file.reader import read_json_from_file
from yta_general_utils.file.writer import write_json_to_file
from yta_general_utils.path import create_file_abspath
from yta_general_utils.file.checker import file_exists
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.downloader.google_drive import GoogleDriveResource
from yta_general_utils.image.color.picker import get_most_common_green_rgb_color_and_similars
from yta_general_utils.image.parser import ImageParser
from yta_general_utils.image.region import ImageRegionFinder
from yta_general_utils.image.converter import pil_image_to_numpy
from moviepy.editor import VideoFileClip, ImageClip, CompositeVideoClip, concatenate_videoclips, vfx
from typing import Union, Any


# TODO: Review all this code please, a lot is duplicated
# TODO: Make this below expect a type 'Image', but 'Image' or
# 'Image.Image' returns each arg must be a type. Got <module 
# 'PIL.Image'. We also can expect np.ndarray when read from
# memory (video.get_frame(...))
def get_greenscreen_areas_details(image: Union[str, Any]):
    """
    This method detects the most common green rgb color and
    the different closed regions in the provided 'image'. The
    'image' must be a filename or a Pillow image instance.

    This method returns a sequence of the 3 found elements:
    the most common green rgb color, the similar green colors
    and the regions found.
    """
    # TODO: Improve the way we check np.ndarray information so
    # we can make sure it is a valid image and not any other
    # kind of information

    # Sometimes it comes from a moviepy clip turned into np
    image = ImageParser.to_pillow(image)
    
    green_rgb_color, similar_greens = get_most_common_green_rgb_color_and_similars(image)
    regions = ImageRegionFinder.find_green_regions(pil_image_to_numpy(image))

    return green_rgb_color, similar_greens, regions


"""
        Moving Greenscreens below
        TODO: Move this below to another more specific file, thank you
        TODO: I'm placing this below here by now, but maybe we should
        build a 'MovingVideoGreenscreen' object for this
"""

# TODO: Move this to a consts.py or something
GREENSCREENS_FOLDER = 'resources/greenscreens/'

def insert_video_into_moving_greescreen(greenscreen_google_drive_url: str, video: Union[VideoFileClip, CompositeVideoClip, ImageClip, str], output_filename: Union[str, None] = None):
    """
    This method inserts the provided 'video' into the moving greenscreen
    that is contained in the provided 'greenscreen_google_drive_url'. It
    will extract all greenscreen frames, analyze each of them and extract
    the greenscreen areas information and put the provided 'video' 
    combined with the greenscreen.

    This method will identify the Google Drive ID to store (or load if
    previously downloaded) the moving greenscreen video to be able to
    handle it. It will use one specific folder to work with the data:
    'resources/greenscreens/GOOGLE_DRIVE_ID', extracting
    the frames, generating a 'config.json' file and also downloading
    the resource from Google Drive.

    @param
        **greenscreen_google_drive_url**
        The Google Drive url of the moving greenscreen video resource.

    @param
        **video**
        The filename of the video we want to put into the greenscreen.

    @param
        **output_filename**
        The filename in which we want to write the final video that is
        our 'video' inside of the greenscreen.
    """
    video = VideoParser.to_moviepy(video)
    
    __preprocess_moving_greenscreen(greenscreen_google_drive_url)

    # Here we have the moving greenscreen with the source file and
    # the config file ready to use
    google_drive_id = GoogleDriveResource(greenscreen_google_drive_url).id
    moving_gs_folder = GREENSCREENS_FOLDER + google_drive_id
    greenscreen_filename = moving_gs_folder + '/greenscreen.mp4'
    greenscreen_config_filename = moving_gs_folder + '/config.json'

    moving_gs = VideoFileClip(greenscreen_filename)

    if moving_gs.duration < video.duration:
        video = video.set_duration(moving_gs.duration)
    elif video.duration < moving_gs.duration:
        moving_gs = moving_gs.set_duration(video.duration)

    # We dynamically load main video frames
    frames = VideoFrameExtractor.get_all_frames(video)

    greenscreen_details = read_json_from_file(greenscreen_config_filename)

    # Now lets put together greenscreen and main video frames
    # applying a mask to build the final video by concatenating
    # them
    final_clips = []
    for index, frame in enumerate(frames):
        frame_details = greenscreen_details[str(index)]

        # We need to put the image into the greenscreen frame
        imageclip = ImageClip(frame, duration = 1 / 60)
        
        # TODO: Pay atenttion to this:
        # I have a 'to_video' method in ImageGreenscreen that
        # should be used here, but I can't. Why? Because of its 
        # behaviour right now. ImageGreenscreen object is not
        # prepared to have numpy images stored to be used later,
        # so when I call the 'to_video' method I should be able
        # to use the frame I get here below, but I can't because
        # it is expecting an image on Google Drive or a local
        # storaged filename. I don't want to write the frame
        # because of performance decreases, so I leave this text
        # here for the future, to make own custom Greenscreen
        # objects able to handle images and videos loaded in
        # memory to work with.

        # Inject the imageclip
        # TODO: This is a sensitive part, documentation says something
        # about masks when numpy provided, but this is working
        green_screen_clip = ImageClip(get_frame_from_video_by_frame_number(moving_gs, index), duration = imageclip.duration).fx(vfx.mask_color, color = frame_details['rgb_color'], thr = 100, s = 5)
        
        # We cannot resize with numpys
        imageclip = imageclip.resize(width = frame_details['lower_right_pixel'][0] - frame_details['upper_left_pixel'][0]).set_position((frame_details['upper_left_pixel'][0], frame_details['upper_left_pixel'][1]))

        final_clips.append(CompositeVideoClip([imageclip, green_screen_clip], size = green_screen_clip.size))

    final = concatenate_videoclips(final_clips).set_audio(video.audio)

    if output_filename:
        final.write_videofile(output_filename, fps = video.fps)

    return final

def __prepare_moving_greenscreen(google_drive_url: str):
    """
    This method will download (if needed) the resource from Google
    Drive to local and will create (if it doesn't exist) the config
    file of a moving Greenscreen.

    TODO: What about other greenscreens (?)

    @param
        **google_drive_url**
        The Google Drive url of the greenscreen. This will be used
        to download the source file and also to identify it with
        the Google Drive ID.
    """
    if not google_drive_url:
        raise Exception('No "google_drive_url" provided.')

    # The identifier of the greenscreen is the Google Drive id
    google_drive_id = GoogleDriveResource(google_drive_url).id
    greenscreen_filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4'
    config_filename = GREENSCREENS_FOLDER + google_drive_id + '/config.json'

    # Check if folder exist or not, and create it if not
    create_file_abspath(get_project_abspath() + greenscreen_filename)

    # TODO: What about image Greenscreen, not .mp4 (?)

    # Check if video exist
    if not file_exists(greenscreen_filename):
        get_resource(google_drive_url, GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4')

    # Check if config file exist
    if not file_exists(config_filename):
        write_json_to_file({}, config_filename)

def __preprocess_moving_greenscreen(greenscreen_google_drive_url: str):
    """
    This method analyzes, frame by frame, a video greenscreen and stores
    each frame information about greenscreen area in a config file to
    be lately used for video injection.
    """
    if not greenscreen_google_drive_url:
        raise Exception('No "greenscreen_google_drive_url" provided.')
    
    # First, make sure config file and source file are available
    __prepare_moving_greenscreen(greenscreen_google_drive_url)

    google_drive_id = GoogleDriveResource(greenscreen_google_drive_url).id
    filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4'

    # Read the video and extract frames
    gs_clip = VideoFileClip(filename)

    for i in range((int) (gs_clip.fps * gs_clip.duration)):
        __process_moving_greenscreen_frame(greenscreen_google_drive_url, i)

    return

def __process_moving_greenscreen_frame(greenscreen_google_drive_url: str, frame_number: int, do_force: bool = False):
    """
    This method processes the 'frame_number' frame of the existing
    greenscreen in 'greenscreen_google_drive_url' source. It will
    detect the greenscreen area of that frame and store it in the
    specific configuration file.

    This method will store the information in an specific folder
    'resources/greenscreens/frames/GOOGLE_DRIVE_ID/'. All
    the information is the greenscreen video, the frames and the
    'config.json' file that contains the detailed greenscreen
    areas of each frame. As you can see, the 'GOOGLE_DRIVE_ID' is
    used as identifier.

    If that 'frame_number' has been previously analyzed, the 
    system will do it again only if the 'do_force' parameter
    is True. If not, it will stop and return.

    @param
        **greenscreen_google_drive_url**
        The Google Drive url that contains the greenscreen video
        from which we will analyze the frame.
    
    @param
        **frame_number**
        The frame number that will be analyzed. The first frame of
        the greenscreen is the number 0.

    @param
        **do_force**
        The information of that 'frame_number' could be stored 
        previously so it would be ignored. You can set this
        parameter to True to force recalculating the frame.
    """
    if not greenscreen_google_drive_url:
        return None
    
    if frame_number < 0:
        return None
    
    # We always make sure greenscreen source file and config are ready
    __prepare_moving_greenscreen(greenscreen_google_drive_url)

    # sample url: https://drive.google.com/file/d/1My5V8gKcXDQGGKB9ARAsuP9MIJm0DQOK/view?usp=sharing
    google_drive_id = GoogleDriveResource(greenscreen_google_drive_url).id
    
    config_filename = GREENSCREENS_FOLDER + google_drive_id + '/config.json'
    greenscreen_filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4'

    json_data = read_json_from_file(config_filename)

    if str(frame_number) in json_data and not do_force:
        # Frame information found and no forcing
        return

    # We make sure we have the clip (and resource)
    gs_clip = VideoFileClip(greenscreen_filename)

    greenscreen_frames_number = (int) (gs_clip.fps * gs_clip.duration)
    if frame_number > (greenscreen_frames_number):
        raise Exception('The requested "frame_number" is not valid. The greenscreen only has "' + str(greenscreen_frames_number) + '" frames.')
    
    # This below stores the frame and use it later for analysis (working)
    # output_filename = create_temp_filename('tmp_frame.png')
    # get_frame_from_video_by_frame_number(gs_clip, frame_number, output_filename)
    # details = get_greenscreen_areas_details(output_filename)

    # This below tries to do with numpy
    # TODO: Maybe this is making it slow (too much memory)
    green_rgb_color, similar_greens, green_regions = get_greenscreen_areas_details(VideoFrameExtractor.get_frame_by_frame_number(gs_clip, frame_number))

    json_data[frame_number] = {
        'index': frame_number,
        'rgb_color': green_rgb_color,
        'similar_greens': similar_greens,
        # TODO: I keep doing this to break not the functionality but
        # we need to be able to handle different green areas now that
        # we are able to do it
        'upper_left_pixel': green_regions[0]['top_left'],
        'lower_right_pixel': green_regions[0]['bottom_right'],
        'frames': None
    }

    return write_json_to_file(json_data, config_filename)


from yta_multimedia.video.utils import generate_videoclip_from_image
from yta_multimedia.resources.image.drive_urls import GOOGLE_SEARCH_IMAGE_GOOGLE_DRIVE_DOWNLOAD_URL
from yta_multimedia.audio.sound.generation.sound_generator import SoundGenerator
from yta_multimedia.resources import get_resource
from yta_multimedia.video.edition.effect.constants import EFFECTS_RESOURCES_FOLDER
from moviepy.editor import TextClip, CompositeVideoClip


class GoogleSearch:
    def generate(self, text: str):
        _EXTENDED_DURATION = 0.5
        _DURATION = 3

        # Download the resource we need
        TMP_FILENAME = get_resource(GOOGLE_SEARCH_IMAGE_GOOGLE_DRIVE_DOWNLOAD_URL, EFFECTS_RESOURCES_FOLDER + 'images/google_search.png')
        clip = generate_videoclip_from_image(TMP_FILENAME, _DURATION + _EXTENDED_DURATION)

        # Calculate each char duration and set texts according to this
        text_len = len(text)
        each_char_duration = _DURATION / text_len

        clips = []
        for i in range(text_len):
            # Generate a text clip for each text writing part
            txt_clip = TextClip(text[:i + 1], font = 'Arial', fontsize = 40, color = 'black')
            clip_duration = each_char_duration
            if i == (text_len - 1):
                clip_duration = each_char_duration + _EXTENDED_DURATION
            
            txt_clip = txt_clip.set_position([380, 465]).set_duration(clip_duration).set_start(i * each_char_duration)

            clips.append(txt_clip)

        video = CompositeVideoClip([clip] + clips)
        # Here we have the text being written, we need the sound
        audio = SoundGenerator.create_typing_audio()
        video = video.set_audio(audio)

        return video


from yta_general_utils.programming.env import get_current_project_env
from yta_general_utils.downloader.image import download_image
from random import randint

import time
import requests


PRODIA_API_KEY =  get_current_project_env('PRODIA_API_KEY')

def generate_image_with_prodia(prompt: str, output_filename: str):
    """
    Generates a new prodia image with the provided 'prompt' and stores it
    locally with the provided 'output_filename'.
    """
    # url = 'https://api.prodia.com/v1/sdxl/generate'

    # Landscape ar = 1344x768 | Vertical ar = 768x1344

    # This below returns a response containing the job working on navigator 
    # and allows me using other models as generators
    # I can inspect here (https://app.prodia.com/) and see model strings
    # https://api.prodia.com/generate?new=true&prompt=a+woman&model=absolutereality_v181.safetensors+%5B3d9d4d2b%5D&negative_prompt=&steps=20&cfg=7&seed=2328045384&sampler=DPM%2B%2B+2M+Karras&aspect_ratio=square
    # {"job":"acffa9a4-6d59-44c0-9f37-bdf1fc5da1c6","status":"queued","params":{"type":"TextToImage","options":{"sd_model_checkpoint":"absolutereality_v181.safetensors [3d9d4d2b]"},"request":{"prompt":"a woman","cfg_scale":7,"steps":20,"negative_prompt":"","seed":2328045384,"sampler_name":"DPM++ 2M Karras","width":512,"height":512}}}

    # payload = {
    #     'new': True,
    #     'prompt': prompt,
    #     #'model': 'absolutereality_v181.safetensors [3d9d4d2b]',   # this model works on above request, not here
    #     'model': 'sd_xl_base_1.0.safetensors [be9edd61]',
    #     #'negative_prompt': '',
    #     'steps': 20,
    #     'cfg_scale': 7,
    #     'seed': 2328045384,
    #     'sampler': 'DPM++ 2M Karras',
    #     'width': 1344,
    #     'height': 768
    # }
    # headers = {
    #     "accept": "application/json",
    #     "content-type": "application/json",
    #     "X-Prodia-Key": PRODIA_API_KEY
    # }
    if not prompt:
        return None
    
    if not output_filename:
        return None

    # If you comment this and uncomment the one below it works
    # seed = randint(1000000000, 9999999999)
    # response = requests.get('https://api.prodia.com/generate?new=true&prompt=' + prompt + '&model=absolutereality_v181.safetensors+%5B3d9d4d2b%5D&steps=20&cfg=7&seed=' + str(seed) + '&sampler=DPM%2B%2B+2M+Karras&aspect_ratio=square')
    payload = {
        'new': True,
        'prompt': prompt,
        #'model': 'absolutereality_v181.safetensors [3d9d4d2b]',   # this model works on above request, not here
        'model': 'sd_xl_base_1.0.safetensors [be9edd61]',
        #'negative_prompt': '',
        'steps': 20,
        'cfg_scale': 7,
        'seed': 2328045384,
        'sampler': 'DPM++ 2M Karras',
        'width': 1344,
        'height': 768
    }
    headers = {
        "accept": "application/json",
        "content-type": "application/json",
        "X-Prodia-Key": PRODIA_API_KEY
    }
    url = 'https://api.prodia.com/v1/sdxl/generate'
    response = requests.post(url, json = payload, headers = headers)
    response = response.json()

    # When requested it is queued, so we ask for it until it is done
    if "status" in response and response['status'] == 'queued':
        job_id = response['job']
        __retrieve_job(job_id, output_filename)
    else:
        print(response)

def __retrieve_job(job_id, output_filename):
    """
    Makes a request for the image that is being generated with the
    provided 'job_id'.

    It has a loop to wait until it is done. This code is critic.
    """
    url = "https://api.prodia.com/v1/job/" + str(job_id)

    headers = {
        "accept": "application/json",
        "X-Prodia-Key": PRODIA_API_KEY
    }

    response = requests.get(url, headers = headers)
    response = response.json()
    #print(response)

    # TODO: Do a passive waiting
    is_downloadable = True

    if response['status'] != 'succeeded':
        is_downloadable = False

    # TODO: Implement a tries number
    while not is_downloadable:
        time.sleep(5)
        print('Doing a request in loop')

        # We do the call again
        response = requests.get(url, headers = headers)
        response = response.json()
        print(response)
        if 'imageUrl' in response:
            is_downloadable = True

    image_url = response['imageUrl']

    return download_image(image_url, output_filename)


from yta_general_utils.checker.type import variable_is_type
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.file.writer import write_file
from moviepy.editor import VideoFileClip, ImageClip
from yta_multimedia.video.parser import VideoParser
from typing import Union
from subprocess import run
import numpy as np


# TODO: Remove, refactor or improve tis method below
def generate_videoclip_from_image(image_filename: Union[ImageClip, str], duration: float = 1, output_filename: Union[str, None] = None):
    """
    Receives an image as 'image_filename' and creates an ImageClip of
    'duration' seconds. It will be also stored as a file if 
    'output_filename' is provided.

    # TODO: Should this method go into 'video.utils' instead of here (?)
    """
    if not image_filename:
        return None
    
    if duration <= 0:
        return None
    
    if not duration:
        return None
    
    if variable_is_type(output_filename, str):
        if not output_filename:
            return None
    
    if variable_is_type(image_filename, str):
        # ADV: By now we are limiting this to 60 fps
        image_filename = ImageClip(image_filename).set_fps(60).set_duration(duration)

    if output_filename:
        image_filename.write_videofile(output_filename)

    return image_filename

# TODO: Move this to another file (maybe 'concatenation.py') (?)
def concatenate_videos_ffmpeg(videos_abspaths, output_abspath: str):
    """
    This method concatenates the videos provided in 'videos_abspaths'
    and builds a new video, stored locally as 'output_abspath'.

    This method uses ffmpeg to concatenate the videos, so they must
    have the same resolution.
    """
    text = ''
    for video_abspath in videos_abspaths:
        text += f'file \'{video_abspath}\'\n'

    filename = create_temp_filename('append_videos.txt')
    write_file(text, filename)

    # TODO: Make a custom call from python not as command
    command = 'ffmpeg -y -f concat -safe 0 -i ' + filename + ' -c copy ' + output_abspath
    run(command)
    
    return VideoFileClip(output_abspath)

def is_video_transparent(video):
    """
    Checks if the first frame of the mask of the given 'video'
    has, at least, one transparent pixel.
    """
    # We need to detect the transparency from the mask
    video = VideoParser.to_moviepy(video, has_mask = True)

    # We need to find, by now, at least one transparent pixel
    # TODO: I would need to check all frames to be sure of this above
    return np.any(video.mask.get_frame(t = 0) == 1)

"""
ChatGPT told me this:

pip install ffmpeg-python

import ffmpeg

# Define tu archivo de entrada y salida
filename = 'tu_archivo.txt'  # Cambia esto al nombre de tu archivo de lista
output_abspath = 'salida.mp4'  # Cambia esto al nombre de tu archivo de salida

# Ejecuta el comando con ffmpeg-python
ffmpeg.input(filename, format='concat', safe='0').output(output_abspath, c='copy').run(overwrite_output=True)
"""


from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from yta_multimedia.greenscreen.custom.utils import get_greenscreen_details
from yta_multimedia.alphascreen.masked_clip_creator import MaskedClipCreator
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.duration import set_video_duration, ExtendVideoMode
from yta_multimedia.greenscreen.enums import GreenscreenType
from yta_general_utils.checker.type import variable_is_type
from yta_general_utils.image.parser import ImageParser
from typing import Union
from moviepy.editor import ImageClip, CompositeVideoClip, vfx, VideoClip
from PIL import Image, ImageDraw
from typing import Union


class ImageGreenscreen:
    greenscreen: GreenscreenDetails = None

    def __init__(self, greenscreen: Union[GreenscreenDetails, str]):
        if variable_is_type(greenscreen, str):
            # We need to automatically detect greenscreen details
            greenscreen = get_greenscreen_details(greenscreen, GreenscreenType.IMAGE)

        self.greenscreen = greenscreen

        # TODO: Do this here to be able to use it in the masked_clip_creator
        TMP_FILENAME = self.greenscreen.get_filename()
        # I consider the same greenscreen rgb color for all areas
        # Duration will be set at the end
        greenscreen_clip = ImageClip(TMP_FILENAME, duration = 1 / 60).fx(vfx.mask_color, color = self.greenscreen.greenscreen_areas[0].rgb_color, thr = 100, s = 5)

        regions = [gsa.region for gsa in self.greenscreen.greenscreen_areas]

        self.masked_clip_creator = MaskedClipCreator(regions, greenscreen_clip)

    def __process_elements_and_save(self, output_filename):
        """
        Processes the greenscreen by writing the title, description
        and any other available element, and stores it locally as
        'output_filename' once processed.
        """
        base = Image.open(self.greenscreen.filename_or_google_drive_url)
        draw = ImageDraw.Draw(base)

        # TODO: I preserve this code for the future
        # # We need to write title if existing
        # if self.__title:
        #     title_position = (self.__title_x, self.__title_y)
        #     draw.text(title_position, self.__title, font = self.__title_font, fill = self.__title_color)

        # if self.__description:
        #     description_position = (self.__description_x, self.__description_y)
        #     draw.text(description_position, self.__description, font = self.__description_font, fill = self.__description_color)

        # TODO: Handle anything else here

        # We save the image
        base.save(output_filename, quality = 100)

    def from_image_to_image(self, image, output_filename: Union[str, None] = None):
        """
        Receives an 'image', places it into the greenscreen and generates
        another image that is stored locally as 'output_filename' if
        provided.
        """
        # TODO: This is not returning RGBA only RGB
        return self.masked_clip_creator.from_image_to_image(image, output_filename)

    def from_images_to_image(self, images, output_filename: str = None):
        return self.masked_clip_creator.from_images_to_image(images, output_filename)

    def from_image_to_video(self, image, duration: float, output_filename: Union[str, None] = None):
        """
        Receives an 'image', places it into the greenscreen and generates
        a video of 'duration' seconds of duration that is returned. This method
        will store locally the video if 'output_filename' is provided.
        """
        return self.masked_clip_creator.from_image_to_video(image, duration, output_filename)

    def from_images_to_video(self, images, duration: float, output_filename: str = None):
        return self.masked_clip_creator.from_images_to_video(images, duration, output_filename)
    
    def from_video_to_video(self, video: Union[str, VideoClip], output_filename: str = None):
        """
        Inserts the provided 'video' in the greenscreen and returns the
        CompositeVideoClip that has been created. If 'output_filename' 
        provided, it will be written locally with that file name.

        The provided 'video' can be a filename or a moviepy video clip.
        """
        return self.masked_clip_creator.from_video_to_video(video, output_filename)
    
    def from_videos_to_video(self, videos: list[Union[str, VideoClip]], output_filename: str = None):
        return self.masked_clip_creator.from_videos_to_video(videos, output_filename)


MANIM_RESOURCES_FOLDER = 'resources/manim/'

# Manim core
MANDATORY_CONFIG_PARAMETER = 1
OPTIONAL_CONFIG_PARAMETER = 0
# I obtained manim dimensions from here: https://docs.manim.community/en/stable/faq/general.html#what-are-the-default-measurements-for-manim-s-scene
SCENE_HEIGHT = 8 # 1080 equivalent
SCENE_WIDTH = 14 + (2 / 9) # 1920 equivalent
HALF_SCENE_WIDTH = SCENE_WIDTH / 2
HALF_SCENE_HEIGHT = SCENE_HEIGHT / 2
LEFT_MARGIN = -HALF_SCENE_WIDTH
UP_MARGIN = HALF_SCENE_HEIGHT




from yta_general_utils.file.writer import write_binary_file

import requests


def narrate_tetyys(text: str, output_filename: str):
    """
    This method creates an audio voice narration of the provided
    'text' read with tthe tetyys system voice (Microsoft Speech
    API 4.0 from 1998) and stores it as 'output_filename'. It is 
    only available for ENGLISH speaking.

    You can change some voice parameters in code to make it a
    different voice.

    This method is requesting an external (but apparently stable
    website).
    """
    # This was taken from here (https://www.tetyys.com/SAPI4/)
    if not text:
        return None
    
    if not output_filename:
        return None
    
    headers = {
        'accept': '*/*',
        'accept-language': 'es-ES,es;q=0.9',
        'priority': 'u=1, i',
        'referer': 'https://www.tetyys.com/SAPI4/',
        'sec-ch-ua': '"Not/A)Brand";v="8", "Chromium";v="126", "Google Chrome";v="126"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36',
    }

    params = {
        'text': text,
        # Inspect options 'value' from https://www.tetyys.com/SAPI4/ but
        # each voice has a pre-set 'pitch' and 'speed' 
        'voice': 'Sam', 
        'pitch': '100',
        'speed': '150',
    }

    """
    Some VOICE options:
        'Male Whisper' 113, 140
        'Female Whisper' 169, 140
        'Mary' 169, 140
        'Mary in Space'|'Mary in Hall'|'Mary in Stadium'|Mary (for Telephone) 169, 140
        'Mike in Space'|... 113, 140
        'RobosoftOne'|'RobosoftTwo'
        'Sam' 100, 140
    """

    response = requests.get('https://www.tetyys.com/SAPI4/SAPI4', params = params, headers = headers)

    write_binary_file(response.content, output_filename)

    return output_filename


from yta_multimedia.video.generation.manim.constants import HALF_SCENE_HEIGHT, HALF_SCENE_WIDTH
from yta_multimedia.video.generation.manim.utils.dimensions import width_to_manim_width, height_to_manim_height
from random import random


def get_random_position(width: float, height: float):
    """
    Returns a random position inside the screen according to the provided element width and
    height to fit in. If you are trying to position a text inside screen limits, you must
    provide text width and height to let this method calculate that random position.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_upper_left_position(width: float, height: float):
    """
    Returns a random position in the upper left corner according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = -(width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_upper_right_position(width: float, height: float):
    """
    Returns a random position in the upper right corner according to the 
    provided element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_upper_center_position(width: float, height: float):
    """
    Returns a random position in the upper center according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH / 2 + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH / 2 - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_upper_position(width: float, height: float):
    """
    Returns a random position in the upper section according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_center_position(width: float, height: float):
    """
    Returns a random position in the center according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH / 2 + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH / 2 - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT / 2 + (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT / 2 - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_center_left_position(width: float, height: float):
    """
    Returns a random position in the center left according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = -(width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT / 2 + (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT / 2 - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_center_right_position(width: float, height: float):
    """
    Returns a random position in the center right according to the 
    provided element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT / 2 + (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT / 2 - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_lower_left_position(width: float, height: float):
    """
    Returns a random position in the lower left corner according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = -(width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = -(height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_lower_right_position(width: float, height: float):
    """
    Returns a random position in the lower right corner according to the 
    provided element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = -(height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_lower_center_position(width: float, height: float):
    """
    Returns a random position in the lower center according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH / 2 + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH / 2 - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = -(height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_lower_position(width: float, height: float):
    """
    Returns a random position in the upper section according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = -(height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

# Exact places
HEIGHT_DISTANCE_FROM_EDGES = height_to_manim_height(10)
WIDTH_DISTANCE_FROM_EDGES = width_to_manim_width(10)
def get_upper_left_position(width: float, height: float):
    """
    Returns the exact position of the upper left corner according to
    the provided element 'width' and 'height' to fit in and be placed
    just in the corner.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = -HALF_SCENE_WIDTH + WIDTH_DISTANCE_FROM_EDGES + (width / 2)
    Y = HALF_SCENE_HEIGHT - HEIGHT_DISTANCE_FROM_EDGES - (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_upper_right_position(width: float, height: float):
    """
    Returns the exact position of the upper right corner according to
    the provided element 'width' and 'height' to fit in and be placed
    just in the corner.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = HALF_SCENE_WIDTH - WIDTH_DISTANCE_FROM_EDGES - (width / 2)
    Y = HALF_SCENE_HEIGHT - HEIGHT_DISTANCE_FROM_EDGES - (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_upper_center_position(height: float):
    """
    Returns the exact position of the upper center position according 
    to the provided element 'width' and 'height' to fit in and be 
    placed just there.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = 0
    Y = HALF_SCENE_HEIGHT - HEIGHT_DISTANCE_FROM_EDGES - (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_left_position(width: float):
    """
    Returns the exact position of the left side according to
    the provided element 'width' and 'height' to fit in and be placed
    just in that place

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = -HALF_SCENE_WIDTH + WIDTH_DISTANCE_FROM_EDGES + (width / 2)
    Y = 0

    return {
        'x': X,
        'y': Y
    }

def get_center_position():
    """
    Returns the exact position of the center according to the 
    provided element 'width' and 'height' to fit in and be placed
    just in that place

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = 0
    Y = 0

    return {
        'x': X,
        'y': Y
    }

def get_right_position(width: float):
    """
    Returns the exact position of the right side according to
    the provided element 'width' and 'height' to fit in and be placed
    just in that place

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = HALF_SCENE_WIDTH - WIDTH_DISTANCE_FROM_EDGES - (width / 2)
    Y = 0

    return {
        'x': X,
        'y': Y
    }

def get_lower_left_position(width: float, height: float):
    """
    Returns the exact position of the lower left corner according to
    the provided element 'width' and 'height' to fit in and be placed
    just in the corner.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = -HALF_SCENE_WIDTH + WIDTH_DISTANCE_FROM_EDGES + (width / 2)
    Y = -HALF_SCENE_HEIGHT + HEIGHT_DISTANCE_FROM_EDGES + (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_lower_right_position(width: float, height: float):
    """
    Returns the exact position of the lower right corner according to
    the provided element 'width' and 'height' to fit in and be placed
    just in the corner.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = HALF_SCENE_WIDTH - WIDTH_DISTANCE_FROM_EDGES - (width / 2)
    Y = -HALF_SCENE_HEIGHT + HEIGHT_DISTANCE_FROM_EDGES + (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_lower_center_position(height: float):
    """
    Returns the exact position of the lower center position according 
    to the provided element 'width' and 'height' to fit in and be 
    placed just there.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = 0
    Y = -HALF_SCENE_HEIGHT + HEIGHT_DISTANCE_FROM_EDGES + (height / 2)

    return {
        'x': X,
        'y': Y
    }


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames import VideoFrameExtractor
from yta_general_utils.file.checker import file_is_video_file
from yta_general_utils.file import list
from yta_general_utils.temp import create_custom_temp_filename 
from moviepy.editor import VideoFileClip, ImageClip, concatenate_videoclips, CompositeVideoClip
from typing import Union

import numpy as np
import cv2


def invert_video_mask(video: Union[str, VideoFileClip]):
    """
    This method will invert the provided video mask. Make 
    sure the video you provide has a mask. This method 
    will return a mask containing the original video
    with the mask inverted.

    This method will iterate over each frame and will
    invert the numpy array to invert the opacity by 
    chaging the zeros by ones and the ones by zeros (but
    in 0 - 255 range).

    This is useful to overlay videos that have alpha 
    channels so they can let see the video behind through
    that alpha channel. For example, if your video is 
    just a manim animation with only text in the middle,
    the other pixels in the video will be alpha, so if you
    invert them you will obtain the text-transparent 
    effect that is a pretty artistic one.

    This mask has to be used in a specific way, that is in
    a CompositeVideoClip, in second position, with the
    'use_bgclip = True' flag to be over the other video.
    """
    video = VideoParser.to_moviepy(video, has_mask = True)

    # TODO: Confirm that this exist or create it if not
    FRAMES_FOLDER_NAME = 'frames_main'
    # This is to avoid memory limit exceeded
    in_memory = False

    # We will invert all frames
    # TODO: Please, try to do this with ffmpeg concatenate
    # images with alpha channel to video because it will be
    # faster but I couldn't in the past
    clips = []
    for i in range(int(video.fps * video.duration)):
        # TODO: Build my own utils method (?)
        mask_frame = VideoFrameExtractor.get_frame_by_frame_number(video.mask, i)
        #mask_frame = video.mask.get_frame(i / 60)
        frame = VideoFrameExtractor.get_frame_by_frame_number(video, i)

        # Invert the mask
        where_0 = np.where(mask_frame == 0)
        where_1 = np.where(mask_frame == 1)
        mask_frame[where_0] = 255
        mask_frame[where_1] = 0

        # Combine the fourth (alpha) channel
        mask_frame = mask_frame[:, :, np.newaxis]
        frame_rgba = np.concatenate((frame, mask_frame), axis = 2)
        if in_memory:
            clips.append(ImageClip(frame_rgba, duration = 1 / 60))
        else:
            tmp_frame_name = create_custom_temp_filename(FRAMES_FOLDER_NAME + '/frame' + str(i).zfill(5) + '.png')
            cv2.imwrite(tmp_frame_name, frame_rgba)
    
    if not in_memory:
        frames_folder = create_custom_temp_filename(FRAMES_FOLDER_NAME + '/')
        images = list(frames_folder, pattern = '*.png')
        for image in images:
            clips.append(ImageClip(image, duration = 1 / 60))

    mask = concatenate_videoclips(clips)

    return mask

# TODO: I think this will be deleted in the future, when refactored
# and when we confirm that here is another way of handling this
def apply_inverted_mask(video: Union[str, VideoFileClip], mask_video: Union[str, VideoFileClip], output_filename: Union[str, None] = None):
    """
    Applies the provided 'mask_video' with its mask inverted
    over the also provided 'video'. This is useful to make
    artistic effects. This methods applies the 
    'invert_video_mask' method to the provided 'mask_video'.
    """
    if not video:
        raise Exception('No "video" provided.')
    
    if not mask_video:
        raise Exception('No "mask_video" provided.')
    
    if isinstance(video, str):
        if not file_is_video_file:
            raise Exception('Provided "video" is not a valid video file.')
        
        video = VideoFileClip(video, has_mask = True)

    if isinstance(mask_video, str):
        if not file_is_video_file:
            raise Exception('Provided "mask_video" is not a valid video file.')
        
        mask_video = VideoFileClip(mask_video, has_mask = True)

    if not mask_video.mask:
        mask_video = VideoFileClip(mask_video.filename, has_mask = True)

    mask_video = invert_video_mask(mask_video)
    # TODO: Handle durations
    final_clip = CompositeVideoClip([video, mask_video.subclip(0, video.duration)], use_bgclip = True)
    final_clip = final_clip.set_audio(video.audio)

    if output_filename:
        final_clip.write_videofile(output_filename)

    return final_clip

# TODO: Look for the way to store the video with the inverted mask locally
# so I can use that video as usual. Normally I download video with alpha
# layers from the Internet or I create them with manim, but now I have to
# handle it manually, so this is a challenge


from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_center import CoordinateCenter
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_corner import CoordinateCorner
from yta_multimedia.video.edition.effect.moviepy.position.enums import ScreenPosition
from yta_general_utils.file.checker import file_is_video_file
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip
from typing import Union


def position_video_in(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], background_video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], position: Union[ScreenPosition, CoordinateCenter, CoordinateCorner]):
    """
    Returns the 'video' positioned (with '.set_position(...)') to stay in 
    the provided 'position' without movement. It won't set any other
    property more than the duration (you will need to manually add
    '.set_duration()' or '.set_start()' if needed).

    This method will return the video positioned as a single element, so 
    make to wrap it properly in an array if it is part of a complex
    animation. 
    """
    if not video:
        raise Exception('No "video" provided.')
    
    if isinstance(video, str):
        if not file_is_video_file:
            raise Exception('Provided "video" is not a valid video file.')
        
        video = VideoFileClip(video)

    if not background_video:
        raise Exception('No "background_video" provided.')

    if isinstance(background_video, str):
        if not file_is_video_file:
            raise Exception('Provided "background_video" is not a valid video file.')
        
        background_video = VideoFileClip(background_video)

    if not isinstance(position, ScreenPosition):
        if not isinstance(position, tuple) and len(position) != 2:
            raise Exception('Provided "position" is not a valid ScreenPosition enum or (x, y) tuple.')
        
    position = get_moviepy_position(video, background_video, position)

    return video.set_position(position)


"""
    Coords related functions below
"""
def get_moviepy_position(video, background_video, position: Union[ScreenPosition, CoordinateCenter, CoordinateCorner]):
    """
    In the process of overlaying and moving the provided 'video' over
    the also provided 'background_video', this method calculates the
    (x, y) tuple position that would be, hypothetically, adapted from
    a 1920x1080 black color background static image. The provided 
    'position' will be transformed into the (x, y) tuple according
    to our own definitions in which the video (that starts in upper left
    corner) needs to be placed to fit the desired 'position'.
    """
    # TODO: Add 'video' and 'background_video' checkings
    if not video:
        raise Exception('No "video" provided.')
    
    if not background_video:
        raise Exception('No "background_video" provided.')
    
    if not position:
        raise Exception('No "position" provided.')
    
    if not isinstance(position, ScreenPosition) and not isinstance(position, CoordinateCenter) and not isinstance(position, CoordinateCorner):
        raise Exception('Provided "position" is not ScreenPosition, CoordinateCenter or CoordinateCorner.')
    
    position_tuple = None

    if isinstance(position, ScreenPosition):
        position_tuple = position.get_moviepy_position(video, background_video)
    elif isinstance(position, CoordinateCenter):
        position_tuple = position.recalculate_for_video(video, background_video)
    elif isinstance(position, CoordinateCorner):
        position_tuple = position.recalculate_for_video(background_video)

    return position_tuple


from yta_multimedia.video.generation.manim.constants import SCENE_HEIGHT, SCENE_WIDTH
from manim import *


def width_to_manim_width(width):
    """
    You provide a real width in pixels (maybe 1920) and it is returned in the
    manim equivalent width. We consider a 16:9 proportion and 1920 as the
    maximum valid width.

    This method is built by myself to work better with a 16:9 proportion of 
    1920x1080 pixels. The manim system is different, so I made this to
    simplify the process.
    """
    return (width * SCENE_WIDTH) / 1920

def manim_width_to_width(width):
    """
    TODO: Write help
    """
    return (width * 1920) / SCENE_WIDTH

def height_to_manim_height(height):
    """
    You provide a real height in pixels (maybe 1080) and it is returned in the
    manim equivalent pixels. We consider 16:9 proportion and 1080 as the 
    maximum valid height.

    This method is built by myself to work better with a 16:9 proportion of 
    1920x1080 pixels. The manim system is different, so I made this to
    simplify the process.
    """
    return (height * SCENE_HEIGHT) / 1080

def manim_height_to_height(height):
    """
    TODO: Write help
    """
    return (height * 1080) / SCENE_HEIGHT




def fitting_text(text, width_to_fit: float = 1920, fill_opacity: float = 1, stroke_width: float = 0, color: ParsableManimColor = None, font_size: float = DEFAULT_FONT_SIZE, line_spacing: float = -1, font: str = '', slant: str = NORMAL, weight: str = NORMAL, t2c: dict[str, str] = None, t2f: dict[str, str] = None, t2g: dict[str, tuple] = None, t2s: dict[str, str] = None, t2w: dict[str, str] = None, gradient: tuple = None, tab_width: int = 4, warn_missing_font: bool = True, height: float = None, width: float = None, should_center: bool = True, disable_ligatures: bool = False, **kwargs):
    """
    This method returns a Text mobject that fits the provided 'width_to_fit'
    or, if the height is greater than the scene height, returns one with the
    greates possible width.
    
    This method has been built to be sure that your text is completely shown
    between the screen margins.

    @param
        **width_to_fit**
        The widht you want to fit, in normal pixels (1920 is the maximum). 
        These pixels will be processed to manim dimensions.
    """
    width_to_fit = width_to_manim_width(width_to_fit)

    txt_width_fitted = Text(text, fill_opacity, stroke_width, color, font_size, line_spacing, font, slant, weight, t2c, t2f, t2g, t2s, t2w, gradient, tab_width, warn_missing_font, height, width, should_center, disable_ligatures, **kwargs).scale_to_fit_width(width_to_fit)
    # I use a margin of 100 pixels so avoid being just in the borders
    txt_height_fitted = Text(text, fill_opacity, stroke_width, color, font_size, line_spacing, font, slant, weight, t2c, t2f, t2g, t2s, t2w, gradient, tab_width, warn_missing_font, height, width, should_center, disable_ligatures, **kwargs).scale_to_fit_height(SCENE_HEIGHT - height_to_manim_height(100))

    # As it is a 16:9 proportion, the height is the measure that limits the most
    if txt_height_fitted.font_size < txt_width_fitted.font_size:
        return txt_height_fitted
    return txt_width_fitted

def fitting_image(filename, width_to_fit, image_mode: str = 'RGBA', **kwargs):
    """
    Returns an ImageMobject of the provided 'filename' image that fits the provided 'width_to_fit'
    or, if the height limit is surpassed, that fits the height limit.

    @param
        **width_to_fit**
        The widht you want to fit, in normal pixels (1920 is the scene 
        width). These pixels will be processed to manim dimensions.
    """
    width_to_fit = width_to_manim_width(width_to_fit)

    image_width_fitted = ImageMobject(filename, image_mode, **kwargs).scale_to_fit_width(width_to_fit)
    image_height_fitted = ImageMobject(filename, image_mode, **kwargs).scale_to_fit_height(width_to_fit)

    # As it is a 16:9 proportion, the height is the measure that limits the most
    if image_height_fitted.width < image_width_fitted.width:
        return image_height_fitted
    
    return image_width_fitted

def fullscreen_image(filename, image_mode: str = 'RGBA', **kwargs):
    """
    Returns an ImageMobject that fits whole screen dimensions (1920x1080). It
    will ignore the dimension that is out of bounds.
    """
    image_width_fitted = ImageMobject(filename, image_mode, **kwargs).scale_to_fit_width(width_to_manim_width(1920))

    # We want the image that occupies the whole screen
    if manim_height_to_height(image_width_fitted.height) >= 1080:
        return image_width_fitted
    
    image_height_fitted = ImageMobject(filename, image_mode, **kwargs).scale_to_fit_height(height_to_manim_height(1080))

    return image_height_fitted

def preprocess_image(image: ImageMobject):
    """
    This method processes images bigger than our 1920x1080 dimensions and returns it
    scaled down to fit those dimensions. You should use this method as the first one
    when working with ImageMobjects, and then scaling it down as much as you need.
    """
    if manim_width_to_width(image.width) > 1920:
        image.scale_to_fit_width(width_to_manim_width(1920))
    if manim_height_to_height(image.height) > 1080:
        image.scale_to_fit_height(height_to_manim_height(1080))

    return image


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, VideoClip
from typing import Union


class FlipVerticallyVideoEffect(VideoEffect):
    """
    This effect flips the video vertically.
    """

    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        Applies the effect to the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        effect_name = VideoEffect.get_moviepy_vfx_effect('mirror_y')
        parameters = {}

        return video.fx(effect_name, **parameters)


from yta_multimedia.video.generation.rrss.discord.classes.discord_conversation_character import DiscordConversationCharacter
from yta_multimedia.image.generation.rrss.discord.discord_message_image_generator import DiscordMessageImageGenerator
from yta_multimedia.video.generation.rrss.discord.constants import DISCORD_MESSAGE_RECEIVED_AUDIO_FILENAME, DISCORD_MESSAGE_RECEIVED_AUDIO_GOOGLE_DRIVE_URL
from yta_multimedia.video.generation.rrss.discord.enums import DiscordConversationMessageMode
from yta_multimedia.resources import get_resource
from yta_general_utils.image.converter import pil_image_to_numpy
from moviepy.editor import AudioFileClip, ColorClip, CompositeVideoClip, ImageClip


class DiscordConversationMessage:
    character = None
    text = None
    duration = None

    def __init__(self, character: DiscordConversationCharacter, text: str, duration: float = None, mode: DiscordConversationMessageMode = DiscordConversationMessageMode.NORMAL, image_url: str = None, sound = None, effect = None):
        # TODO: Ignore 'sound' and 'effect' by now. Implement later
        if not character:
            raise Exception('No "character" provided.')
        
        # TODO: Check that character is 'DiscordConversationCharacter' type

        if not text:
            raise Exception('No "text" provided.')
        
        # TODO: Check if 'mode' is 'DiscordConversationMessageMode' type
        if not mode:
            mode = DiscordConversationMessageMode.NORMAL

        # TODO: Check if 'image_url' and that is valid image url

        # TODO: Check if 'sound' that is valid

        # TODO: Check if 'effect' that is valid

        if not duration:
            # No fixed duration so we calculate it (default: SLOW mode)
            character_duration = 0.10
            self.duration = 0.6
            if mode == DiscordConversationMessageMode.NORMAL:
                self.duration = 0.45
                character_duration = 0.08
            elif mode == DiscordConversationMessageMode.FAST:
                self.duration = 0.3
                character_duration = 0.06

            if len(text) > 8:
                # Long messages need longer duration
                self.duration = character_duration * len(text)
        else:
            self.duration = duration

        self.character = character
        self.text = text
        # TODO: Handle 'image_url'
        # TODO: Handle 'effect'
        self.discord_message_image_generator = DiscordMessageImageGenerator()

    def generate_clip(self):
        """
        Generates a clip with the discord message received audio, the
        specific 'sound' if provided, with the discord message in the
        center of the video.

        This method returns a moviepy CompositeVideoClip.
        """
        audioclip = AudioFileClip(get_resource(DISCORD_MESSAGE_RECEIVED_AUDIO_GOOGLE_DRIVE_URL, DISCORD_MESSAGE_RECEIVED_AUDIO_FILENAME))
        # TODO: Handle 'self.sound' and use CompositeAudioClip

        image = self.discord_message_image_generator.generate(self.character.username, self.character.avatar_url, self.text)
        # Original image is too wide, we shorten it by the half 
        # TODO: We could do a x3 and limit the amount of chars in text
        image = image.crop((0, 0, image.width / 2, image.height))
        image = image.resize((image.width * 2, image.height * 2))
        # Open it with Image.open(image)

        # TODO: Make this duration dynamic
        clip = CompositeVideoClip([
            # Black background to make the Image fit
            ColorClip((1920, 1080), (0, 0, 0), duration = self.duration),
            # We set it in the center of the video
            ImageClip(pil_image_to_numpy(image), duration = self.duration).set_position((1920 / 2 - image.width / 2, 1080 / 2 - image.height / 2))
        ])

        if audioclip.duration > clip.duration:
            audioclip.subclip(0, clip.duration)

        clip = clip.set_audio(audioclip).set_fps(60)

        return clip


from yta_general_utils.file.checker import file_is_audio_file
from yta_general_utils.temp import create_temp_filename
from moviepy.editor import AudioFileClip, CompositeAudioClip
from pydub import AudioSegment
from typing import Union

import numpy as np
import io
import scipy.io.wavfile as wavfile


# TODO: Review, improve, test and refactor all these methods below
# By detecting the instance we can avoid 'audiosegment_to_X' because
# we know the origin is AudioSegment so we just need 'to_x'
class AudioParser:
    def to_audiosegment(cls, audio):
        """
        Forces the provided 'audio' to be a pydub AudioSegment
        and returns it if valid 'audio' provided or raises an
        Exception if not.
        """
        if not audio:
            raise Exception('No "audio" provided.')
        
        if isinstance(audio, str):
            if not file_is_audio_file(audio):
                raise Exception('Provided "audio" filename is not a valid audio file.')
            audio = AudioSegment.from_file(audio)
        elif isinstance(audio, np.ndarray):
            # TODO: Check this
            audio = cls.numpy_to_audiosegment(audio)
        elif isinstance(audio, (AudioFileClip, CompositeAudioClip)):
            audio = cls.moviepy_audio_to_audiosegment(audio)

        return audio

    @classmethod
    def to_audiofileclip(cls, audio):
        """
        Forces the provided 'audio' to be a moviepy AudioFileClip
        and returns it if valid 'audio' provided or raises an
        Exception if not.
        """
        if not audio:
            raise Exception('No "audio" provided.')
        
        if isinstance(audio, str):
            if not file_is_audio_file(audio):
                raise Exception('Provided "audio" filename is not a valid audio file.')
            audio = AudioFileClip(audio)
        elif isinstance(audio, np.ndarray):
            # TODO: Check this works
            audio = cls.numpy_to_audiofileclip(audio)
        elif isinstance(audio, AudioSegment):
            audio = cls.audiosegment_to_audiofileclip(audio)

        return audio

    @classmethod
    def audiosegment_to_audiofileclip(cls, audio: AudioSegment):
        """
        This method reads the provided AudioSegment as a buffer and converts
        it to a moviepy AudioFileClip without writting any file.

        TODO: Please, make it through memory and not writting files.
        """
        if not audio:
            raise Exception('No "audio" provided.')
        
        if not isinstance(audio, AudioSegment):
            raise Exception('The "audio" parameter provided is not an AudioSegment.')
        
        # TODO: I have not been able to create an AudioFileClip dinamically
        # from memory information. I don't want to write but...
        tmp_filename = create_temp_filename('tmp_audio.wav')
        audio.export(tmp_filename, format = 'wav')

        return AudioFileClip(tmp_filename)

    @classmethod
    def moviepy_audio_to_audiosegment(cls, audio: Union[AudioFileClip, CompositeAudioClip]):
        """
        This method returns the provided moviepy audio converted into a
        pydub AudioSegment.

        TODO: This method currently writes a temporary file to make the 
        conversion. This needs to be improved to avoid writting files.
        """
        if not audio:
            raise Exception('No "audio" provided.')
        
        if not isinstance(audio, AudioFileClip):
            raise Exception('The "audio" parameter provided is not an AudioFileClip.')

        # TODO: Please, improve this to be not writting files
        tmp_filename = create_temp_filename('tmp_audio.wav')
        audio.write_audiofile(tmp_filename)
        audio = AudioSegment.from_file(tmp_filename, format = 'wav')

        return audio

    @classmethod
    # TODO: This has not been tested properly
    def numpy_to_audiosegment(cls, audio, sample_rate):
        """
        Convers the provided 'audio' numpy array, that contains the audio data
        and must be in float32 or int16, to a pydub AudioSegment.
        """
        # Normalize audio_array if it's not already in int16 format
        if audio.dtype != np.int16:
            if audio.dtype != np.float32:
                raise Exception('The "audio" parameter provided is not np.int16 nor np.float32.')
            
            # Assuming the audio_array is in float32 with values between -1 and 1
            audio = (audio * 32767).astype(np.int16)
        
        # Create a BytesIO object
        with io.BytesIO() as buffer:
            wavfile.write(buffer, sample_rate, audio)
            buffer.seek(0)
            
            audio_segment = AudioSegment.from_file(buffer, format = 'wav')
        
        return audio_segment

    @classmethod
    # TODO: This method has not been tested properly.
    def audiosegment_to_numpy(cls, audio: AudioSegment):
        """
        This method turns the provided 'audio' AudioSegment into a numpy
        array by converting it first to an AudioFileClip and then to a
        numpy.

        TODO: Please, maybe it is a better (more direct) way

        TODO: This method has not been tested properly
        """
        if not audio:
            raise Exception('No "audio" provided.')
        
        if not isinstance(audio, AudioSegment):
            raise Exception('The "audio" parameter provided is not an AudioSegment.')
        
        # TODO: Maybe this is not the best way, I need to test and improve this
        return np.array(audio.get_array_of_samples())
        #return audiofileclip_to_numpy(audiosegment_to_audiofileclip(audio))

    @classmethod
    # TODO: This has not been tested properly
    def audiofileclip_to_numpy(cls, audio: Union[AudioFileClip]):
        """
        Convers the provided 'audio' moviepy AudioFileClip to a numpy
        array that will be np.float32.
        """
        if not audio:
            raise Exception('No "audio" provided.')
        
        if not isinstance(audio, AudioFileClip):
            raise Exception('The "audio" parameter provided is not a moviepy AudioFileClip.')

        chunk_size = 5*1024*1024
        audio_chunks = []
        for chunk in audio.iter_chunks(chunksize = chunk_size):
            # Convertir cada fragmento a un array numpy y añadirlo a la lista
            audio_array = np.frombuffer(chunk, dtype=np.int16)
            
            # Normalizar si el audio es estéreo (tendría dos columnas)
            if len(audio_array) > 0 and len(audio_array) % 2 == 0:
                audio_array = audio_array.reshape(-1, 2)
            
            audio_chunks.append(audio_array)
        
        # Concatenar todos los fragmentos en un solo array
        full_audio_array = np.concatenate(audio_chunks, axis = 0)
        
        # Convertir a float32 y normalizar
        full_audio_array = full_audio_array.astype(np.float32)
        if np.max(np.abs(full_audio_array)) > 1.0:
            full_audio_array /= np.max(np.abs(full_audio_array))
        
        return full_audio_array

    @classmethod
    # TODO: This has not been tested properly
    def numpy_to_audiofileclip(cls, audio: np.ndarray):
        if not audio:
            raise Exception('No "audio" provided.')
        
        if not isinstance(audio, np.ndarray):
            # TODO: Check this works
            raise Exception('The "audio parameter provided is not a np.ndarray.')
        
        # TODO: This won't work
        return AudioFileClip(audio)


# TODO: This custom green screens need work, refactor and 
# make them work better and simplify te code.
# Consider this part as experimental.
# Was working in the first stage but it is not clear how 
# to handle it in code (x.x)

# I kept working on them and I simplified the process.
# We can add some image or video into an image or
# video greenscreen. We can customize any greenscreen
# but we will have them stored online